@@ -132,12 +132,23 @@ That said it is your responsiblity to use the right policy for your particular c
 """
 
 from contextlib import suppress
-from functools import partial, wraps
-from collections import Counter, defaultdict
+from functools import partial, wraps, cached_property
+from collections import defaultdict
 
 from dataclasses import dataclass, field
-from typing import Callable, MutableMapping, Sized, Union, Optional, Iterable, Any
+from itertools import chain
+from operator import attrgetter
+from typing import (
+    Callable,
+    MutableMapping,
+    Union,
+    Optional,
+    Iterable,
+    Any,
+    Mapping,
+)
 
+from i2 import double_up_as_factory
 from i2.signatures import (
     call_somewhat_forgivingly,
     Parameter,
@@ -151,6 +162,8 @@ from meshed.base import (
     _mk_func_nodes,
     _func_nodes_to_graph_dict,
     is_func_node,
+    FuncNodeAble,
+    func_node_transformer,
 )
 
 from meshed.util import (
@@ -159,6 +172,11 @@ from meshed.util import (
     NotUniqueError,
     NotFound,
     NameValidationError,
+    Renamer,
+    _if_none_return_input,
+    numbered_suffix_renamer,
+    replace_item_in_iterable,
+    InvalidFunctionParameters,
 )
 from meshed.itools import (
     topological_sort,
@@ -532,7 +550,7 @@ class DAG:
 
     """
 
-    func_nodes: Iterable[Union[FuncNode, Callable]]
+    func_nodes: Iterable[Union[FuncNode, Callable]] = ()
     cache_last_scope: bool = field(default=True, repr=False)
     parameter_merge: ParameterMerger = field(
         default=conservative_parameter_merge, repr=False
@@ -821,8 +839,10 @@ class DAG:
             if isinstance(obj, str):
                 obj = obj.split()
             if isinstance(obj, (str, Callable)):
+                # TODO: See if we can use _func_node_for instead
                 return [self.get_node_matching(obj)]
             elif isinstance(obj, Iterable):
+                # TODO: See if we can use _func_node_for instead
                 return list(map(self.get_node_matching, obj))
             else:
                 raise ValidationError(f'Unrecognized variables specification: {obj}')
@@ -850,6 +870,74 @@ class DAG:
             func, self.func_nodes, lambda func, fn: func == fn.func
         )
 
+    # TODO: Reflect: Should we include functions as keys here? Makes existence of the
+    #  item depend on unicity of the function in the DAG, therefore dynamic, so instable?
+    @cached_property
+    def _func_node_for(self):
+        """A dictionary mapping identifiers and functions to their FuncNode instances
+        in the DAG. The keys of this dictionary will include:
+
+        - identifiers (names) of the ``FuncNode`` instances
+        - ``out`` of ``FuncNode`` instances
+        - The ``.func`` of the ``FuncNode`` instances if it's unique.
+
+        >>> def foo(x): return x + 1
+        >>> def bar(x): return x * 2
+        >>> dag = DAG([
+        ...     FuncNode(foo, out='foo_output'),
+        ...     FuncNode(bar, name='B', out='b', bind={'x': 'foo_output'}),
+        ... ])
+
+        A ``FuncNode`` instance is indexed by both its identifier (``.name``) as well as
+        the identifier of it's output (``.out``):
+
+        >>> dag._func_node_for['foo_output']
+        FuncNode(x -> foo -> foo_output)
+        >>> dag._func_node_for['foo']
+        FuncNode(x -> foo -> foo_output)
+        >>> dag._func_node_for['b']
+        FuncNode(x=foo_output -> B -> b)
+        >>> dag._func_node_for['B']
+        FuncNode(x=foo_output -> B -> b)
+
+        If the function is hashable (most are) and unique within the ``DAG``, you
+        can also find the ``FuncNode`` via the ``.func`` it's wrapping:
+
+        >>> dag._func_node_for[foo]
+        FuncNode(x -> foo -> foo_output)
+        >>> dag._func_node_for[bar]
+        FuncNode(x=foo_output -> B -> b)
+
+        A word of warning though: The function index is provided as a convenience, but
+        using identifiers is preferable since referencing via the function object
+        depends on the other functions of the DAG, so could change if we add nodes.
+        
+        """
+        d = dict()
+        for func_node in self.func_nodes:
+            d[func_node.out] = func_node
+            d[func_node.name] = func_node
+
+            try:
+                if func_node.func not in d:
+                    # if .func not in d already, remember the link
+                    d[func_node.func] = func_node
+                else:
+                    # if .func was already in there, mark it for removal
+                    # (but leaving the key present so that we know about the duplication)
+                    d[func_node.func] = None
+            except TypeError:
+                # ignore (and don't include func) if not hashable
+                pass
+
+        # remove the items marked for removal and return
+        return {k: v for k, v in d.items() if v is not None}
+
+    def find_func_node(self, node, default=None):
+        if isinstance(node, FuncNode):
+            return node
+        return self._func_node_for.get(node, default)
+
     def __iter__(self):
         """Yields the self.func_nodes
         Note: The raison d'etre of this ``__iter__`` is simply because if no custom one
@@ -923,10 +1011,27 @@ class DAG:
                 # ``parameter_merge`` with raise an error.)
                 yield self.parameter_merge(params_with_name_changed_to_src_name)
 
+    # TODO: Find more representative (and possibly shorter) doctest:
     @property
     def graph_ids(self):
         """The dict representing the ``{from_node: to_nodes}`` graph.
         Like ``.graph``, but with node ids (names).
+
+        >>> from meshed.dag import DAG
+        >>> def add(a, b=1): return a + b
+        >>> def mult(x, y=3): return x * y
+        >>> def exp(mult, a): return mult ** a
+        >>> assert DAG([add, mult, exp]).graph_ids == {
+        ...     'a': ['add_', 'exp_'],
+        ...     'b': ['add_'],
+        ...     'add_': ['add'],
+        ...     'x': ['mult_'],
+        ...     'y': ['mult_'],
+        ...     'mult_': ['mult'],
+        ...     'mult': ['exp_'],
+        ...     'exp_': ['exp']
+        ... }
+
         """
         return {
             _name_attr_or_x(k): list(map(_name_attr_or_x, v))
@@ -951,6 +1056,147 @@ class DAG:
             other = list(DAG(other).func_nodes)
         return DAG(list(self.func_nodes) + other)
 
+    def copy(self, renamer=numbered_suffix_renamer):
+        return DAG(ch_names(self.func_nodes, renamer=renamer))
+
+    def add_edge(self, from_node, to_node, to_param=None):
+        """Add an e
+
+        :param from_node:
+        :param to_node:
+        :param to_param:
+        :return: A new DAG with the edge added
+
+        >>> def f(a, b): return a + b
+        >>> def g(c, d=1): return c * d
+        >>> def h(x, y=1): return x ** y
+        >>>
+        >>> three_funcs = DAG([f, g, h])
+        >>> assert (
+        ...     three_funcs(x=1, c=2, a=3, b=4)
+        ...     == (1, 2, 7)
+        ...     == (h(x=1), g(c=2), f(a=3, b=4))
+        ...     == (1 ** 1, 2 * 1, 3 + 4)
+        ... )
+        >>> print(three_funcs.synopsis_string())
+        x,y -> h_ -> h
+        c,d -> g_ -> g
+        a,b -> f_ -> f
+        >>> hg = three_funcs.add_edge('h', 'g')
+        >>> assert (
+        ...     hg(a=3, b=4, x=1)
+        ...     == (7, 1)
+        ...     == (f(a=3, b=4), g(c=h(x=1)))
+        ...     == (3 + 4, 1 * (1 ** 1))
+        ... )
+        >>> print(hg.synopsis_string())
+        a,b -> f_ -> f
+        x,y -> h_ -> h
+        h,d -> g_ -> g
+        >>>
+        >>> fhg = three_funcs.add_edge('h', 'g').add_edge('f', 'h')
+        >>> assert (
+        ...     fhg(a=3, b=4)
+        ...     == 7
+        ...     == g(h(f(3, 4)))
+        ...     == ((3 + 4) * 1) ** 1
+        ... )
+        >>> print(fhg.synopsis_string())
+        a,b -> f_ -> f
+        f,y -> h_ -> h
+        h,d -> g_ -> g
+
+        The from and to nodes can be expressed by the ``FuncNode`` ``name`` (identifier)
+        or ``out``, or even the function itself if it's used only once in the ``DAG``.
+
+        >>> fhg = three_funcs.add_edge(h, 'g').add_edge('f_', 'h')
+        >>> assert fhg(a=3, b=4) == 7
+
+        By default, the edge will be added from ``from_node.out`` to the first
+        parameter of the function of ``to_node``.
+        But if you want otherwise, you can specify the parameter the edge should be
+        connected to.
+        For example, see below how we connect the outputs of ``g`` and ``h`` to the
+        parameters ``a`` and ``b`` of ``f`` respectively:
+
+        >>> f_of_g_and_h = (
+        ...     DAG([f, g, h])
+        ...     .add_edge(g, f, to_param='a')
+        ...     .add_edge(h, f, 'b')
+        ... )
+        >>> assert (
+        ...     f_of_g_and_h(x=2, c=3, y=2, d=2)
+        ...     == 10
+        ...     == f(g(c=3, d=2), h(x=2, y=2))
+        ...     == 3 * 2 + 2 ** 2
+        ... )
+        >>>
+        >>> print(f_of_g_and_h.synopsis_string())
+        x,y -> h_ -> h
+        c,d -> g_ -> g
+        g,h -> f_ -> f
+
+        See Also ``DAG.add_edges`` to add multiple edges at once
+
+        """
+        # resolve from_node and to_node into FuncNodes
+        from_node, to_node = map(self.find_func_node, (from_node, to_node))
+        if to_node is None and callable(to_node):
+            to_node = FuncNode(
+                to_node
+            )  # TODO: Automatically avoid clashing with dag identifiers (?)
+
+        # if to_param is None, take the first parameter of to_node as the one
+        if to_param is None:
+            if not to_node.bind:
+                raise InvalidFunctionParameters(
+                    "You can't add an edge TO a FuncNode whose function has no "
+                    'parameters. '
+                    f'You attempted to add an edge between {from_node=} and {to_node=}.'
+                )
+            else:
+                # first param of .func (i.e. first key of .bind)
+                to_param = next(iter(to_node.bind))
+
+        existing_bind = to_node.bind[to_param]
+        if any(existing_bind == fn.out for fn in self.func_nodes):
+            raise ValueError(
+                f"The {to_node} node is already sourcing '{to_param}' from '"
+                f"{existing_bind}'."
+                'Delete that edge to be able before you add a new one'
+            )
+
+        new_to_node_dict = to_node.to_dict()
+        new_bind = new_to_node_dict['bind'].copy()
+        new_bind[to_param] = from_node.out  # this is the actual edge creation
+        new_to_node = FuncNode.from_dict(dict(new_to_node_dict, bind=new_bind))
+        return DAG(
+            replace_item_in_iterable(
+                self.func_nodes,
+                condition=lambda x: x == to_node,
+                replacement=lambda x: new_to_node,
+            )
+        )
+
+    # TODO: There are optimization and pre-validation opportunities here!
+    def add_edges(self, edges):
+        """Adds multiple edges by applying ``DAG.add_edge`` multiple times.
+
+        :param edges: An iterable of ``(from_node, to_node)`` pairs or
+            ``(from_node, to_node, param)`` triples.
+        :return: A new dag with the said edges added.
+
+        >>> def f(a, b): return a + b
+        >>> def g(c, d=1): return c * d
+        >>> def h(x, y=1): return x ** y
+        >>> fhg = DAG([f, g, h]).add_edges([(h, 'g'), ('f_', 'h')])
+        >>> assert fhg(a=3, b=4) == 7
+        """
+        dag = self
+        for edge in edges:
+            dag = dag.add_edge(*edge)
+        return dag
+
     # ------------ display --------------------------------------------------------------
 
     def synopsis_string(self, bind_info='values'):
@@ -973,6 +1219,8 @@ class DAG:
         # >>> assert list(DAG(func_nodes).dot_digraph_body()) == [
         # ]
         """
+        if isinstance(start_lines, str):
+            start_lines = start_lines.split()
         yield from dot_lines_of_func_nodes(self.func_nodes, start_lines=start_lines)
 
     @wraps(dot_digraph_body)
@@ -1177,31 +1425,9 @@ def print_dag_string(dag: DAG, bind_info: str = 'hybrid'):
     print(dag.synopsis_string(bind_info=bind_info))
 
 
-# ---------- with ext.gk -------------------------------------------------------
-
-with suppress(ModuleNotFoundError, ImportError):
-    from meshed.ext.gk import operation, Operation
-
-    def funcs_to_operations(*funcs, exclude_names=()) -> Operation:
-        """Get an operation from a callable"""
-
-        for func in funcs:
-            _func_name = mk_func_name(func, exclude_names)
-            exclude_names = exclude_names + (_func_name,)
-            needs = arg_names(func, _func_name, exclude_names)
-            exclude_names = exclude_names + tuple(needs)
-            yield operation(
-                func, name=_func_name, needs=needs, provides=_func_name,
-            )
-
-    def funcs_to_operators(*funcs, exclude_names=()) -> Operation:
-        """Get an operation from a callable"""
-
-        for func, operation in zip(funcs, funcs_to_operations(funcs, exclude_names)):
-            yield operation(func)
-
-
+# ---------------------------------------------------------------------------------------
 # reordering funcnodes
+
 from meshed.util import uncurry, pairs
 
 mk_mock_funcnode_from_tuple = uncurry(mk_mock_funcnode)
@@ -1221,3 +1447,173 @@ def reorder_on_constraints(funcnodes, outs):
     func_nodes, var_nodes = _separate_func_nodes_and_var_nodes(ordered_nodes)
 
     return func_nodes, var_nodes
+
+
+def attribute_vals(objs: Iterable, attrs: Iterable[str], egress=None):
+    """Extract attributes from an iterable of objects
+    >>> list(attribute_vals([print, map], attrs=['__name__', '__module__']))
+    [('print', 'builtins'), ('map', 'builtins')]
+    """
+    if isinstance(attrs, str):
+        attrs = attrs.split()
+    val_tuples = map(attrgetter(*attrs), objs)
+    if egress:
+        return egress(val_tuples)
+    else:
+        return val_tuples
+
+
+names_and_outs = partial(attribute_vals, attrs=('name', 'out'), egress=chain)
+
+DagAble = Union[DAG, Iterable[FuncNodeAble]]
+
+
+def _validate_func_src(func_src, func_nodes: DagAble):
+    allowed_identifiers = set(names_and_outs(DAG(func_nodes).func_nodes))
+    if not_allowed := (func_src.keys() - allowed_identifiers):
+        raise ValueError(
+            f"These identifiers weren't found in func_nodes: {not_allowed}"
+        )
+    if not_callable := set(filter(lambda x: not callable(x), func_src.values())):
+        raise ValueError(f"These values of func_src weren't callable: {not_callable}")
+
+
+# TODO: extract egress functionality to decorator
+@double_up_as_factory
+def ch_funcs(func_nodes: DagAble = None, *, func_src=(), strict=False):
+    if isinstance(func_nodes, DAG):
+        egress = DAG
+    else:
+        egress = list
+
+    func_src = dict(func_src)
+    if strict:
+        _validate_func_src(func_src, func_nodes)
+
+    def gen():
+        for fn in func_nodes:
+            if (
+                new_func := func_src.get(fn.out, func_src.get(fn.name, None))
+            ) is not None:
+                new_fn_kwargs = dict(fn.to_dict(), func=new_func)
+                yield FuncNode.from_dict(new_fn_kwargs)
+            else:
+                yield fn
+
+    return egress(gen())
+
+
+change_funcs = ch_funcs  # back-compatibility
+
+
+# TODO: extract egress functionality to decorator
+@double_up_as_factory
+def ch_names(func_nodes: DagAble = None, *, renamer: Renamer = numbered_suffix_renamer):
+    """Renames variables and functions of a ``DAG`` or iterable of ``FuncNodes``.
+
+    :param func_nodes: A ``DAG`` of iterable of ``FuncNodes``
+    :param renamer: A function taking an old name and returning the new one, or:
+        - A dictionary ``{old_name: new_name, ...}`` mapping old names to new ones
+        - A string, which will be appended to all identifiers of the ``func_nodes``
+    :return: func_nodes with some or all identifiers changed. If the input ``func_nodes``
+    is an iterable of ``FuncNodes``, a list of func_nodes will be returned, and if the
+    input ``func_nodes`` is a ``DAG`` instance, a ``DAG`` will be returned.
+
+    >>> from meshed.makers import code_to_dag
+    >>> from meshed.dag import print_dag_string
+    >>>
+    >>> @code_to_dag
+    ... def dag():
+    ...     b = f(a)
+    ...     c = g(x=a)
+    ...     d = h(b, y=c)
+    ...
+
+    This is what the dag looks like:
+
+    >>> print_dag_string(dag)
+    x=a -> g -> c
+    a -> f -> b
+    b,y=c -> h -> d
+
+    Now, if rename the vars of the ``dag`` without further specifying how, all of our
+    nodes (names) will be suffixed with a ``_1``
+
+    >>> new_dag = ch_names(dag)
+    >>> print_dag_string(new_dag)
+    a=a_1 -> f_1 -> b_1
+    x=a_1 -> g_1 -> c_1
+    b=b_1,y=c_1 -> h_1 -> d_1
+
+    If any nodes are already suffixed by ``_`` followed by a number, the default
+    renamer (``numbered_suffix_renamer``) will increment that number:
+
+    >>> another_new_data = ch_names(new_dag)
+    >>> print_dag_string(another_new_data)
+    x=a_2 -> g_2 -> c_2
+    a=a_2 -> f_2 -> b_2
+    b=b_2,y=c_2 -> h_2 -> d_2
+
+    If we specify a string for the ``renamer`` argument, it will be used to suffix all
+    the nodes.
+
+    >>> print_dag_string(ch_names(dag, renamer='_copy'))
+    a=a_copy -> f_copy -> b_copy
+    x=a_copy -> g_copy -> c_copy
+    b=b_copy,y=c_copy -> h_copy -> d_copy
+
+    Finally, for full functionality on renaming, you can use a function
+
+    >>> print_dag_string(ch_names(dag, renamer=lambda x: f"{x.upper()}"))
+    a=A -> F -> B
+    x=A -> G -> C
+    b=B,y=C -> H -> D
+
+    In all the above our input was a ``DAG`` so we got a ``DAG`` back, but if we enter
+    an iterable of ``FuncNode`` instances, we'll get a list of the same back.
+    Also, know that if your function returns ``None`` for a given identifier, it will
+    have the effect of not changing that identifier.
+
+    >>> ch_names(dag.func_nodes, renamer=lambda x: x.upper() if x in 'abc' else None)
+    [FuncNode(x=A -> g -> C), FuncNode(a=A -> f -> B), FuncNode(b=B,y=C -> h -> d)]
+
+    If you want to rename the nodes with an explicit mapping, you can do so by
+    specifying this mapping as your renamer
+
+    >>> substitutions = {'a': 'alpha', 'b': 'bravo'}
+    >>> print_dag_string(ch_names(dag, renamer=substitutions))
+    a=alpha -> f -> bravo
+    x=alpha -> g -> c
+    b=bravo,y=c -> h -> d
+
+    """
+    if isinstance(func_nodes, DAG):
+        egress = DAG
+    else:
+        egress = list
+    renamer = renamer or numbered_suffix_renamer
+    if isinstance(renamer, str):
+        suffix = renamer
+        renamer = lambda name: f'{name}{suffix}'
+    elif isinstance(renamer, Mapping):
+        old_to_new_map = dict(renamer)
+        renamer = old_to_new_map.get
+    assert callable(renamer), f'Could not be resolved into a callable: {renamer}'
+    ktrans = partial(_rename_node, renamer=renamer)
+    func_node_trans = partial(func_node_transformer, kwargs_transformers=ktrans)
+    return egress(map(func_node_trans, func_nodes))
+
+
+def _rename_node(fn_kwargs, renamer: Renamer = numbered_suffix_renamer):
+    fn_kwargs = fn_kwargs.copy()
+    # decorate renamer so if the original returns None the decorated will return input
+    renamer = _if_none_return_input(renamer)
+    fn_kwargs['name'] = renamer(fn_kwargs['name'])
+    fn_kwargs['out'] = renamer(fn_kwargs['out'])
+    fn_kwargs['bind'] = {
+        param: renamer(var_id) for param, var_id in fn_kwargs['bind'].items()
+    }
+    return fn_kwargs
+
+
+rename_nodes = ch_names  # back-compatibility
